from ..text_cleaner import TextCleaner
import pytest



@pytest.mark.parametrize("bigrams,expected_result",
                         [
                             (('short','sleeve'), True),
                             (('a','sleeve'), False),
                             (('the','sleeve'), False)
                         ])
def test_pos_filter(bigrams, expected_result):
    tc = TextCleaner(word_count_min=1, word_length_min=2, 
                     bigram_kwargs={'bigrams_pmi_min_value':1, 'bigrams_min_freq':20})
    assert tc._pos_filter(bigrams) == expected_result
    

@pytest.mark.parametrize("tokens,expected_result",
                         [
                             (['short','sleeve'], {('short','sleeve'):'short sleeve'}),
                             (['a','sleeve'], {}),
                             (['the','sleeve'], {})
                         ])
def test_create_bigrams_dict(tokens, expected_result):
    tc = TextCleaner(word_count_min=1, word_length_min=2, 
                     bigram_kwargs={'bigrams_pmi_min_value':0, 'bigrams_min_freq':0})
    tc._create_bigrams_dict(tokens)
    assert tc.bigrams_dict == expected_result
    
    
@pytest.mark.parametrize("tokens,expected_result",
                         [
                             ([['short','sleeve']], [['short sleeve']]),
                             ([['a','sleeve']], [[]]),
                             ([['the','sleeve']], [[]])
                         ])
def test_get_bigrams(tokens, expected_result):
    tc = TextCleaner(word_count_min=1, word_length_min=2, 
                     bigram_kwargs={'bigrams_pmi_min_value':0, 'bigrams_min_freq':0})
    assert tc._get_bigrams(tokens, fit=True) == expected_result


def test_preprocess_text_filter_tokens():
    tc = TextCleaner(word_count_min=1, word_length_min=2, bigram_kwargs={})
    documents = ['This is a short, exciting DoCuMeNt!!!', 'This is another.']
    tokenized_documents = tc._preprocess_text(documents, fit=True)
    assert tokenized_documents == [['this','is','a','short','exciting','document'], 
                                   ['this','is','another']]
    cleaned_tokens= tc._filter_tokens(tokenized_documents, fit=True)
    assert cleaned_tokens == [['short','exciting','document'],[]]
    

def test_filter_word_count():
    tc = TextCleaner(word_count_min=2, word_length_min=2, bigram_kwargs={})
    tc.word_counts = {'short':1,'exciting':2,'document':3}
    cleaned_tokens = [['short','exciting','document']]
    filtered_tokens = tc._filter_word_count(cleaned_tokens, fit=False)
    assert filtered_tokens == [['exciting','document']]
    
    
def test_clean_documents_no_bigrams():
    tc = TextCleaner(word_count_min=1, word_length_min=2, bigram_kwargs={})
    documents = ['This is a short, exciting DoCuMeNt!!!', 'This is another.']
    cleaned_tokens = tc._clean_documents(documents, fit=True)
    assert cleaned_tokens == [['short','exciting','document'],[]]
    

def test_clean_documents_yes_bigrams():
    tc = TextCleaner(word_count_min=1, word_length_min=2, 
                     bigram_kwargs={'bigrams_pmi_min_value':0, 'bigrams_min_freq':0})
    documents = ['short sleeve', 'another short sleeve']
    cleaned_tokens = tc._clean_documents(documents, fit=True)
    assert cleaned_tokens == [['short sleeve','short','sleeve'],
                              ['short sleeve','short','sleeve']]


from gensim import corpora, models
import numpy as np
from collections import Counter
import itertools
import pickle
import logging
logger = logging.getLogger()
logging.basicConfig(
    format='%(asctime)s %(levelname)-8s %(message)s',
    level=logging.INFO,
    datefmt='%Y-%m-%d %H:%M:%S'
)
from .text_cleaner import TextCleaner


class TagExtractor:
    """
    The TagExtractor trains an LSI model and uses it to extract tags and rank candidates.
    
    Args:
        word_count_min (int): The minimum number of times a word has to appear to be kept. Optional, default is 2.
        word_length_min (int): The minimum length a word must be to be kept. Optional, default is 2.
        num_lsi_topics (int): The number of LSI topics to extract during model training. 
            Optional, default is 300. Optimal values are known to be 100-400 but I've had great results up to 1,000.
        bigram_kwargs (dict): Bigram arguments. 
            bigrams_pmi_min_value: The minimum value of Pointwise Mutual Information for a bigram to be kept. 
                Values above 1 are best. Optional, default is 1.
            bigrams_min_freq: The minimum number of times a bigram has to appear to be kept. 
                Optional, default is 20.
            Optional, default is {'bigrams_pmi_min_value':1,'bigrams_min_freq':20}.
    """
    
    def __init__(self, 
                 word_count_min=2, 
                 word_length_min=2, 
                 num_lsi_topics=300,
                 bigram_kwargs={}):
        self.word_count_min = word_count_min
        self.word_length_min = word_length_min
        self.num_lsi_topics = num_lsi_topics
        self.bigram_kwargs = bigram_kwargs
        
    def save(self, fname=None):
        """
        Saves current model artifacts in the pickle format.
        
        Args:
            fname (str): File name to save to. Optional, default 'TagExtractor.p'.
            
        Returns:
            None.
        
        Raises:
            None.
        """
        with open('TagExtractor.p' if fname is None else fname, 'wb') as f:
            pickle.dump(self.__dict__, f)
    
    def load(self, fname=None):
        """
        Loads model artifacts from the pickle format.
        
        Args:
            fname (str): File name to load. Optional, default 'TagExtractor.p'.
            
        Returns:
            None.
        
        Raises:
            None.
        """
        self.__dict__.update(
            pickle.loads(
                open('TagExtractor.p' if fname is None else fname, 'rb').read()
            )
        )
    
    def fit(self, documents):
        """
        Fits model to a set of documents.
        
        Args:
            documents (list of strings): Documents to model.
            
        Returns:
            None.
            
        Raises:
            None.
        """
        # Clean text
        self.tc = TextCleaner(word_count_min=self.word_count_min, 
                              word_length_min=self.word_length_min,
                              bigram_kwargs=self.bigram_kwargs)
        cleaned_documents = self.tc.fit_transform(documents)
            
        # Create document lookup
        self.problem_docs = []
        self.doc2ind = {}
        for n, (doc, cleaned_doc) in enumerate(zip(documents, cleaned_documents)):
            if len(cleaned_doc)==0:
                self.problem_docs.append(doc)
            self.doc2ind[doc] = n
                
        # Warn for empty documents
        if len(self.problem_docs) > 0:
            logger.warning("Warning: Some documents yield no clean tokens. These documents won't have tags and are stored in self.problem_docs.")
            logger.warning(f'Examples: {self.problem_docs[:5]}')
        
        # Train TF-IDF
        logger.info('Training TF-IDF...')
        self.dictionary = corpora.Dictionary(cleaned_documents)
        self.corpus = [self.dictionary.doc2bow(doc) for doc in cleaned_documents]
        self.tfidf = models.TfidfModel(self.corpus)
        self.corpus_tfidf = self.tfidf[self.corpus]

        # Train LSI
        logger.info('Training LSI...')
        self.lsi_model = models.LsiModel(self.corpus_tfidf, 
                                         id2word=self.dictionary, 
                                         num_topics=self.num_lsi_topics)
        self.corpus_lsi = self.lsi_model[self.corpus_tfidf]
        
        # Save the topic matrix for tag extraction
        self.lsi_topic_matrix = self.lsi_model.get_topics()
        self.lsi_topic_matrix_T_normed = (
            self.lsi_topic_matrix.T/np.linalg.norm(self.lsi_topic_matrix.T, axis=1)[:,None]
        ).T
        
        logger.info('Tag Extractor training is done!')
        
    def _get_vector_representations(self, document):
        """
        Helper method to convert a document into its TF-IDF and LSI vector counterparts.
        
        Args:
            document (str): Document to convert to vectors.
            
        Raises:
            None.
            
        Returns:
            corpus_tfidf (list of tuples): Tuples containing word IDs and TF-IDF values.
            corpus_lsi (list of tuples): Tuples containing word IDs and LSI values.
        """
        doc_ind = self.doc2ind.get(document) 
        if doc_ind is None:
            cleaned_documents = self.tc.transform([document])
            corpus = self.dictionary.doc2bow(cleaned_documents[0])
            corpus_tfidf = self.tfidf[corpus]
            corpus_lsi = self.lsi_model[corpus_tfidf]
        else:
            corpus_tfidf = self.corpus_tfidf[doc_ind]
            corpus_lsi = self.corpus_lsi[doc_ind]
        return corpus_tfidf, corpus_lsi
    
    def _lsi_corpus2vec(self, lsi_corpus_vec):
        """
        Helper method to convert a Gensim LSI representation to a numpy representation.
        
        Args:
            lsi_corpus_vec (list of tuples): Gensim LSI vector.
            
        Raises:
            None
        
        Returns:
            vec (array): Numpy LSI vector.
        """
        vec = np.zeros(self.num_lsi_topics)
        for ind, value in lsi_corpus_vec:
            vec[ind] = value
        return vec
    
    def _extract_adjacent_tags(self, input_tags, n_adjacent_tags):
        """
        Helper method to extract adjacent tags.
        
        Args:
            input_tags (list of strings): Input, reference tags.
            
        Raises:
            None
            
        Returns:
            adjacent_tags (list of strings): Adjacent tags.
        """
        bow = self.dictionary.doc2bow(input_tags)
        bow_vec = self._lsi_corpus2vec(self.lsi_model[self.tfidf[bow]])
        bow_vec_normed = bow_vec/np.linalg.norm(bow_vec)
        sims = bow_vec_normed.dot(self.lsi_topic_matrix_T_normed)
        top_inds = np.argsort(sims)[::-1]

        adjacent_tags = []
        count = 0
        for t in top_inds:
            tag = self.dictionary[t]
            flag = True
            for input_tag in input_tags:
                if (input_tag in tag) or (tag in input_tag):
                    flag = False
            if flag:
                count += 1
                adjacent_tags.append(tag)
            if count == n_adjacent_tags:
                break
        return adjacent_tags
    
    def transform(self, input_document, candidate_documents,
                  n_input_tags=10, n_candidate_tags=5, n_adjacent_tags=10):
        """
        Transforms pairs of input and candidate documents into their respective tag representations.
        
        Args:
            input_document (str): Document of input item.
            candidate_documents (list of strings): Documents of candidate items.
            n_input_tags (int): Maximum number of most common tags to aggregate associated with the input item.
            n_candidate_tags (int): Maximum of tags to extract from each candidate item.
            n_adjacent_tags (int): Maximum number of adjacent tags to extract, given the input_tags.
            
        Raises:
            None.
            
        Returns:
            input_tags (ordered list of tuples): Tuples of tags and number of times extracted from candidate_documents.
            candidate_tags (ordered list of tuples): Tuples of tags and LSI values.
            adjacent_tags (ordered list of strings): Adjacent tags to input_tags.
        """
        tfidf_input, lsi_input = self._get_vector_representations(input_document)
        if (len(tfidf_input) == 0) | (len(lsi_input) == 0):
            return [], []

        candidate_tags = []
        for candidate_document in candidate_documents:
            tfidf_candidate, lsi_candidate = self._get_vector_representations(candidate_document)

            if (len(tfidf_candidate) == 0) | (len(lsi_candidate) == 0):
                candidate_tags.append({})
            else:

                # Get shared words using the sparse TF-IDF corpus
                shared_word_ids = np.array(list(set(np.array(tfidf_input)[:,0]) \
                                              & set(np.array(tfidf_candidate)[:,0]))).astype(int)

                # Get word-topic vectors for those shared words
                shared_word_topics = self.lsi_topic_matrix.T[shared_word_ids]

                # Sum those word-topic vectors weighed by each document's topic loadings
                shared_lsi_word_loadings = np.sum((shared_word_topics*np.array(lsi_input)[:,1]) \
                                                + (shared_word_topics*np.array(lsi_candidate)[:,1]), axis=1)

                # Rank shared words by highest weighed word-topic loadings
                ranking = np.argsort(shared_lsi_word_loadings)[::-1]
                ranked_shared_word_ids = shared_word_ids[ranking]

                tags = [self.dictionary[r] for r in ranked_shared_word_ids][:n_candidate_tags]
                scores = shared_lsi_word_loadings[ranking][:n_candidate_tags].tolist()

                candidate_tags.append(dict(zip(tags, scores)))

        all_candidate_tags = list(itertools.chain(*[list(ct.keys()) for ct in candidate_tags]))
        input_tags = Counter(all_candidate_tags).most_common(n_input_tags)
        adjacent_tags = []
        if n_adjacent_tags > 0:
            adjacent_tags = self._extract_adjacent_tags([t[0] for t in input_tags], n_adjacent_tags)

        return input_tags, candidate_tags, adjacent_tags
        
    def rank(self, candidate_tags, selected_tags):
        """
        Ranks an ordered list of candidate tags given selected tags.
        
        Args:
            candidate_tags (list of dictionaries): candidate_tags from the transform() method.
                Ordered dictionaries containing candidate tags and associated LSI values.
            selected_tags (list of strings): Tags chosen for re-ranking candidates.
            
        Raises:
            None.
            
        Returns:
            ranking (list of ints): List of indices in which to re-rank candidates.
        """
        inds, scores = [], []
        original_ranking = []
        for n, tag_dict in enumerate(candidate_tags):
            document_score = sum(tag_dict.get(tag,0) for tag in selected_tags)
            if document_score > 0:
                scores.append(document_score)
                inds.append(n)
            else:
                original_ranking.append(n)
        reranking = np.argsort(np.array(scores))[::-1].tolist()
        new_partial_ranking = np.array(inds)[reranking].tolist()
        new_ranking = new_partial_ranking + original_ranking
        return new_ranking


from ..tag_extractor import TagExtractor
import numpy as np



def test_lsi_corpus2vec():
    tc = TagExtractor(num_lsi_topics=5)
    result = tc._lsi_corpus2vec([(0,10),(3,1)])
    assert (result == np.array([10,0,0,1,0])).all()


def test_rank():
    tc = TagExtractor()
    candidate_tags = [{'a':1},{'b':2},{'b':3},{'c':4}]
    selected_tags = ['b']
    new_ranking = tc.rank(candidate_tags, selected_tags)
    assert new_ranking == [2,1,0,3]


import re
from gensim.parsing.preprocessing import STOPWORDS
from collections import defaultdict
import itertools
from tqdm import tqdm
import nltk
nltk.download('averaged_perceptron_tagger')


def _tqdm(iterator, display, desc=None):
    """
    Helper wrapper over tqdm to control displaying progress.
    
    Args:
        iterator (list, generator): Iterator to iterate through.
        display (bool): Whether to display progress.
        desc (str): Explanation for work on the iterator. Optional, default is None.
        
    Raises:
        None.
        
    Returns:
        iterator (list, generator, tqdm object): Iterator with tqdm wrapper if display is True.
    """
    if display:
        return tqdm(iterator, position=0, 
                    leave=True, desc=desc)
    return iterator


class TextCleaner:
    def __init__(self, word_count_min=1, word_length_min=2, bigram_kwargs={}):
        self.punctuation = """!"$%\'()*+,./:;<=>?@[\\]^_`{|}~"""
        self.remove_punctuation_rule = re.compile(f"[{re.escape(self.punctuation)}]")
        self.word_counts = defaultdict(lambda :0)
        self.word_count_min = word_count_min
        self.word_length_min = word_length_min
        self.bigrams = (len(bigram_kwargs)>0)
        if self.bigrams:
            self.bigrams_pmi_min_value = bigram_kwargs.get('bigrams_pmi_min_value', 1)
            self.bigrams_min_freq = bigram_kwargs.get('bigrams_min_freq', 20)
        keep_stopwords = ['top','bottom','back','front','full','her','him','herself',
                          'himself','his','hers','kg','km','cm','thick','thin','under',
                          'you','your','yours']
        self.STOPWORDS = keep_stopwords = [s for s in STOPWORDS if s not in keep_stopwords]
        
    def _pos_filter(self, bigram):
        """
        Helper method to filter bigrams based on part of speech (POS).
        
        Args:
            bigram (tuple): bigram to filter. Example: ('word1','word2').
            
        Raises:
            None.
            
        Returns:
            (bool): True/False whether to keep bigram.
        """
        for word in bigram:
            if (word in self.STOPWORDS) or (len(word)<self.word_length_min):
                return False
        acceptable_types = ('JJ', 'JJR', 'JJS', 'NN', 'NNS', 'NNP', 'NNPS')
        second_type = ('NN', 'NNS', 'NNP', 'NNPS')
        tags = nltk.pos_tag(bigram)
        if (tags[0][1] in acceptable_types) and (tags[1][1] in second_type):
            return True
        else:
            return False

    def _create_bigrams_dict(self, tokens):
        """
        Helper method to create map of bigrams to extract.
        
        Args:
            tokens (list of strings): Ordered list of tokens.
            
        Raises:
            None.
            
        Returns:
            bigrams_dict (dict): Dictionary of bigrams to extract. 
        """
        # Create bigrams
        bigram_finder = nltk.collocations.BigramCollocationFinder.from_words(tokens)

        # Filter bigrams by minimum frequency
        bigram_finder.apply_freq_filter(self.bigrams_min_freq)

        # Calculate PMI
        self.bigrams_dict = list(bigram_finder.score_ngrams(nltk.collocations.BigramAssocMeasures().pmi))

        # Filter bigrams based on POS and min PMI
        self.bigrams_dict = {b[0]:f'{b[0][0]} {b[0][1]}' for b in self.bigrams_dict 
                             if ((b[1]>=self.bigrams_pmi_min_value) and self._pos_filter(b[0]))}
        
    def _get_bigrams(self, tokens, fit):
        """
        Helper method to extract bigrams from tokens.
        
        Args:
            tokens (list of lists of strings): List of documents split into tokens.
            fit (bool): Whether to fit new bigrams or use existing mapping to extract bigrams.
            
        Raises:
            None.
            
        Returns:
            bigram_documents (list of lists of strings): Extracted bigrams from tokens.
        """
        if fit:
            self._create_bigrams_dict(list(itertools.chain(*tokens)))
        bigram_documents = []
        for bigrams in _tqdm((list(zip(words[:-1], words[1:])) for words in tokens), 
                             display=fit, desc='Making bigrams'):
            temp_bigrams = []
            for bigram in bigrams:
                keep_bigram = self.bigrams_dict.get(bigram)
                if keep_bigram is not None:
                    temp_bigrams.append(keep_bigram)
                    if fit:
                        self.word_counts[keep_bigram] += 1
            bigram_documents.append(temp_bigrams)
        return bigram_documents
    
    def _preprocess_text(self, documents, fit):
        """
        Helper method to preprocess text by removing punctuation, extra spaces, and lower casing.
        
        Args:
            documents (list of strings): Documents to preprocess.
            fit (bool): Whether to display progress.
            
        Raises:
            None.
            
        Returns:
            tokens (list of lists of strings): Transformed documents into tokens.
        """
        tokens = []
        for text in _tqdm(documents, display=fit, desc='Preprocessing text'):
            text = self.remove_punctuation_rule.sub(' ', text)
            text = re.sub(' +', ' ', text)
            tokens.append(text.strip().lower().split())
        return tokens

    def _filter_tokens(self, tokens, fit):
        """
        Helper method to filter tokens based on stopwords and word_length_min, and fit word_counts.
        One step downstream of self._preprocess_text.
        
        Args:
            tokens (list of lists of strings): List of documents split into tokens.
            fit (bool): Whether to fit word_counts.
            
        Raises:
            None.
        
        Returns:
            cleaned_tokens (list of lists of strings): Cleaned tokens.
        """
        cleaned_tokens = []
        for token_list in _tqdm(tokens, display=fit, desc='Filtering tokens'):
            temp_tokens = []
            for word in token_list:
                if (word not in self.STOPWORDS) & (len(word) >= self.word_length_min):
                    temp_tokens.append(word)
                    if fit:
                        self.word_counts[word] += 1
            cleaned_tokens.append(temp_tokens)
        return cleaned_tokens
    
    def _filter_word_count(self, cleaned_tokens, fit):
        """
        Helper method to filter tokens based on word_count_min.
        
        Args:
            cleaned_tokens (list of lists of strings): List of documents split into cleaned tokens.
            fit (bool): Whether to display progress.
            
        Raises:
            None.
        
        Returns:
            (list of lists of strings): Filtered tokens.
        """
        return [[word for word in ct if (self.word_counts.get(word, 0) >= self.word_count_min)] 
                for ct in _tqdm(cleaned_tokens, display=fit, desc='Filtering by word count')]

    def _clean_documents(self, documents, fit):
        """
        Helper method to clean documents.
        
        Args:
            documents (list of strings): Documents to clean.
            fit (bool): Whether to fit cleaner and display progress.
            
        Raises:
            None.
            
        Returns:
            cleaned_tokens (list of lists of strings): Cleaned tokens.
        """
        tokens = self._preprocess_text(documents, fit)
        cleaned_tokens = self._filter_tokens(tokens, fit)
        if self.bigrams:
            bigram_tokens = self._get_bigrams(tokens, fit)
            cleaned_tokens = [doc[0]+doc[1] for doc in zip(bigram_tokens, cleaned_tokens)]
        if self.word_count_min <= 1:
            return cleaned_tokens
        else:
            return self._filter_word_count(cleaned_tokens, fit)
        
    def fit_transform(self, documents):
        """
        Fits TextCleaner and transforms documents.
        
        Args:
            documents (list of strings): Documents to fit and transform.
            
        Raises:
            None.
            
        Returns:
            cleaned_tokens (list of lists of strings): Cleaned tokens.
        """
        cleaned_tokens = self._clean_documents(documents, True)
        self.word_counts = dict(self.word_counts)
        return cleaned_tokens
    
    def transform(self, documents):
        """
        Transforms documents given the fit TextCleaner.
        
        Args:
            documents (list of strings): Documents to transform.
            
        Raises:
            None.
            
        Returns:
            (list of lists of strings): Cleaned tokens.
        """
        return self._clean_documents(documents, False)


from setuptools import setup, find_packages


setup(
    name='lsi_tagger',
    version='0.1',
    license='MIT',
    author='David Albrecht',
    author_email='davidpabloalbrecht@gmail.com',
    description='LSI based, pairwise tag extraction intended for e-commerce product descriptions.',
    packages=find_packages(),
    url='https://github.com/dpalbrecht/lsi-tagger',
    keywords=['nlp','e-commerce','keyword-extraction','latent-semantic-indexing'],
    install_requires=['gensim','numpy','nltk']
)